/*
 Copyright 2015 Nicolas Melot

 This file is part of Pelib-SCC.

 Pelib-SCC is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pelib-SCC is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Pelib-SCC. If not, see <http://www.gnu.org/licenses/>.

*/


/*
 * sync.c
 *
 *  Created on: 17 Apr 2012
 *  Copyright 2012 Nicolas Melot
 *
 * This file is part of pelib-scc.
 * 
 *     pelib-scc is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     pelib-scc is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with pelib-scc. If not, see <http://www.gnu.org/licenses/>.
 * 
 */
#include <stdlib.h>

#include <RCCE.h>

#include <pelib/sync.h>
#include <pelib/util.h>

int
pelib_scc_flag_alloc(pelib_scc_flag* flag)
{
  pelib_scc_flag tmp;

//  pelib_scc_printf("[CORE %d] Allocating %u %s (rounded up to %u)\n", RCCE_ue(), sizeof(pelib_scc_flag_content), sizeof(pelib_scc_flag_content) <= 1 ? "byte" : "bytes", pelib_scc_allocatable_up(sizeof(pelib_scc_flag_content)));
  tmp = (pelib_scc_flag)RCCE_malloc(pelib_scc_allocatable_up(sizeof(pelib_scc_flag_content)));

  if(tmp != NULL)
  {
    *flag = tmp;
//	pelib_scc_printf("%X\n", *flag);
    return 1;
  }
  else
  {
    return 0;
  }
}

int
pelib_scc_flag_alloc_local(pelib_scc_flag* flag)
{ 
  *flag = (pelib_scc_flag)malloc(sizeof(pelib_scc_flag));

  if(*flag != NULL)
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

int
pelib_scc_flag_free(pelib_scc_flag* flag)
{
  RCCE_free((volatile unsigned char*)*flag);
  return 1;
}

int
pelib_scc_flag_free_local(pelib_scc_flag* flag)
{
  free((void*)*flag);
  return 1;
}

int
pelib_scc_flag_write(pelib_scc_flag* flag, pelib_scc_flag_status value, int destination)
{
//	pelib_scc_flag_status val;
	int res;
//	pelib_scc_printf("[CORE %d] Writing %d to flag\n", RCCE_ue(), value);
//	pelib_scc_printf("[CORE %d] BEFORE: flag=%u, *flag=%u, **flag=%u, value=%u\n", RCCE_ue(), flag, *flag, **flag, value);

	/*
	pelib_scc_printf("[CORE %d] BEFORE: flag=%u\n", RCCE_ue(), flag);
	pelib_scc_printf(", *flag=%u\n", *flag);
	pelib_scc_printf(", **flag=%u\n", **flag);
	pelib_scc_printf(", value=%u\n", value);
	*/

	res = RCCE_put((volatile unsigned char*)*flag, (volatile unsigned char *)&value, pelib_scc_allocatable_up(sizeof(pelib_scc_flag_content)), destination);
//	pelib_scc_printf("[CORE %d] AFTER: flag=%u, *flag=%u, **flag=%u, value=%u\n", RCCE_ue(), flag, *flag, **flag, value);
	switch(res)
	{
		case RCCE_SUCCESS:
//			pelib_scc_flag_read(*flag, &val, destination);
//			pelib_scc_printf("[CORE %d] Success, written %d on flag to MPB of core %d\n", RCCE_ue(), val, destination);
			return 1;
		break;
		case RCCE_ERROR_TARGET:
			if(*flag != NULL)
	  		{
	  			**flag = value;
//				pelib_scc_printf("[CORE %d] Successully written flag to local memory\n", RCCE_ue());
	  			return 1;
			}
			else
			{
//				pelib_scc_printf("[CORE %d] Failed to write flag to local memory\n", RCCE_ue());
				return 0;
			}	
		break;
		default:
//			pelib_scc_printf("[CORE %d] Failed to write flag to MPB (res=%d; length=%d)\n", RCCE_ue(), res, pelib_scc_allocatable_up(sizeof(volatile unsigned int)));
			return 0;
		break;
	}
}

int
pelib_scc_flag_read(pelib_scc_flag flag, pelib_scc_flag_status *value, int destination)
{
	int res;
	volatile unsigned char tmp[SCC_LINE_SIZE];
//	pelib_scc_printf("[CORE %d] Reading from flag\n", RCCE_ue());
//	pelib_scc_printf("[CORE %d] BEFORE: flag=%u, *flag=%u, value=%u, *value=%u (size: %d)\n", RCCE_ue(), flag, *flag, value, *value, pelib_scc_allocatable_up(sizeof(pelib_scc_flag_content)));
	//memcpy in RCCE_get writes from flag to value and writes beyond space allocated to value and override the content of flag, allocated 4 bytes after value
	res = RCCE_get(tmp, (volatile unsigned char*)flag, pelib_scc_allocatable_up(sizeof(pelib_scc_flag_content)), destination);
	res = RCCE_SUCCESS;
//	pelib_scc_printf("[CORE %d] AFTER: flag=%u, *flag=%u, value=%u, *value=%u\n", RCCE_ue(), flag, *flag, value, *value);
	switch(res)
	{
		case RCCE_SUCCESS:
			*value = *(pelib_scc_flag_status*)tmp;
//			pelib_scc_printf("[CORE %d] Successfully read flag (%u) from MPB\n", RCCE_ue(), *value);
			return 1;
		break;
		case RCCE_ERROR_SOURCE:
			if(flag != NULL)
	  		{
	  			*value = *flag;
//				pelib_scc_printf("[CORE %d] Successfully read flag from local memory\n", RCCE_ue());
	  			return 1;
			}
			else
			{
				return 0;
			}	
		break;
		default:
//			pelib_scc_printf("[CORE %d] Failed to read flag from MPB (res=%d; length=%d)\n", RCCE_ue(), res, pelib_scc_allocatable_up(sizeof(volatile unsigned int)));
			return 0;
		break;
	}
}

int
pelib_scc_wait_until(pelib_scc_flag flag, pelib_scc_flag_status status)
{
        pelib_scc_flag_status read;
	pelib_scc_flag_read(flag, &read, RCCE_ue());

	while(status != read)
	{       
		pelib_scc_flag_read(flag, &read, RCCE_ue());
	}

	return 1;
}

//#define foobar fprintf(stderr, "[CORE %d][%s:%s:%d] foobar\n", RCCE_ue(), __FILE__, __FUNCTION__, __LINE__);
#if !defined _OPENMP || 1
void
pelib_scc_critical_begin()
{
	pelib_scc_flag_status read;
        pelib_scc_flag_status status;

        pelib_scc_flag_alloc(&__pelib_scc_critical_flag);
	pelib_scc_flag_read(__pelib_scc_critical_flag, &read, RCCE_ue());
	//printf("[CORE %d][%s:%s:%d] %d %d\n", RCCE_ue(), __FILE__, __FUNCTION__, __LINE__, read, PELIB_SCC_FLAG_SET);

        status = PELIB_SCC_FLAG_UNSET;
        pelib_scc_flag_write(&__pelib_scc_critical_flag, PELIB_SCC_FLAG_UNSET, RCCE_ue());
        
        RCCE_barrier(&RCCE_COMM_WORLD);
        
        if(RCCE_ue() > 0)
        {
		//fprintf(stderr, "[CORE %d] Waiting for core %i\n", RCCE_ue(), RCCE_ue() - 1);
		
		pelib_scc_wait_until(__pelib_scc_critical_flag, PELIB_SCC_FLAG_SET);
        }

	pelib_scc_flag_read(__pelib_scc_critical_flag, &read, RCCE_ue());
	//printf("[CORE %d][%s:%s:%d] %d %d\n", RCCE_ue(), __FILE__, __FUNCTION__, __LINE__, read, PELIB_SCC_FLAG_SET);
}

void
pelib_scc_critical_end()
{
	if(RCCE_ue() < RCCE_num_ues() - 1)
        {       
		//fprintf(stderr, "[CORE %d] Releasing core %i\n", RCCE_ue(), RCCE_ue() + 1);
                pelib_scc_flag_write(&__pelib_scc_critical_flag, PELIB_SCC_FLAG_SET, RCCE_ue() + 1);
        }
        
        RCCE_barrier(&RCCE_COMM_WORLD);
        pelib_scc_flag_free(&__pelib_scc_critical_flag);
}
#endif
